unit MPDrivePanel;

(* TMPDrivePanel v0.1,

   TMPDrivePanel shows a pane with buttons for each existing drive (also dynamic refreshing is possible)
   customizable look and feel is also available ;-)

   credits to : raymond allan (the author of the famous fmedit9[5|8]) for the idea :-)

   (C) 1999 by markus stephany, merkes@t-online.de, http://merkes.home.pages.de/
   ============all rights reserverd===================================================

*)



interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls,Buttons;

const
// message-wparams for the dynamic refresh (WM_DEVICECHANGE)
  DBT_DEVICEARRIVAL = $8000;            // new device has been added
  DBT_DEVICEREMOVECOMPLETE = $8004;     // device has been removed

type
// eventhandler type for the drive button clicks
  TONClickButton = procedure (Sender,Button:Tobject; const ADrive : Char) of Object;

// THE component
  TMPDrivePanel = class(TCustomPanel)
  private
    FFlat: Boolean; // flat buttons ?
//    FLargeIcons: Boolean; // not used at the moment
    FOnClickButton: TOnClickButton; // button click handler
    FButtonWidth: Integer; // width of the drive buttons
    FButtonHeight: Integer; // height of the drive buttons

    FButtonHintTemplate: string; // template for ech button's hint
    FButtonCaptionTemplate: string; // template for each button's caption
    { note about the templates above:
      each occurance of '[DR]' in the template string will be replaced by the respective drive char (uppercase), e.g. :
        template 'Drive "[DR]:"' will be parsed to 'Drive "C:"' for the button for drive 'c:'
      each occurance of '[DD]' in the template string will be replaced by the shell's drive description, e.g. :
        template 'Drive "[DD]"' will be parsed to 'Drive "3.5 Floppy (A:)"' (or something similar) for the 'a:' button
    }

    FSHIl : THandle; // handle of the shell's image list (for the button glyphs)
    FAlign: TAlign;  // overwritten for automatic horizontal/vertical button order
    FVertical: Boolean; // if set manually, true = each button below the other, false = each button to the right of the other
    FButtonSpacing: Integer; // empty space between each button and the next one, also affects the left/top margin of the buttons
    FAutoSize: Boolean; // if true, Align=alLeft,alRight,alTop,alBottom will automatically set the panels's height or width
    FAllowAllUp: Boolean; // for grouping of the drive buttons
    FGroupIndex: Integer;  // ''
    FDrive: Char; // set or read the current drive (push down the corresponding button on write, is only writeable when groupindex <> 0
    FShowGlyphs: Boolean; // to show or hide the shell like drive symbols on the buttons
    FShowCaptions: Boolean; // to show or hide the buttons' caption (parsed from FButtonCaptionTemplate)
    procedure SetFlat(const Value: Boolean);
//    procedure SetLargeIcons(const Value: Boolean);
    procedure SetButtonWidth(const Value: Integer);
    procedure SetButtonHeight(const Value: Integer);
    procedure SetButtonHintTemplate(const Value: string);
    procedure SetButtonCaptionTemplate(const Value: string);
    procedure SetAlign(const Value: TAlign);
    procedure SetVertical(const Value: Boolean);
    procedure SetButtonSpacing(const Value: Integer);
    procedure SetAutoSize(const Value: Boolean);
    procedure SetAllowAllUp(const Value: Boolean);
    procedure SetGroupIndex(const Value: Integer);
    procedure SetDrive(const Value: Char);
    function GetDrive:Char;
    procedure SetShowCaptions(const Value: Boolean);
    procedure SetShowGlyphs(const Value: Boolean);
    procedure DoRefresh;
    { Private declarations }
  protected
    { Protected declarations }
    procedure Loaded ; Override;
    procedure DoBtnClick (Sender : TObject);dynamic;
    procedure CreateWnd ; override;
    function ParseDS ( const aDrive: Char; const Templ: string): string;
  public
    { Public declarations }
    procedure RefreshButtons; // to manually check the buttons for existance of the underlaying drives
    constructor Create (AOwner : TComponent) ; override;
    destructor Destroy; override;

    function HandleDeviceChange(var Msg:TMessage) : Boolean; // use this for automatic refresh
    { note about the automatic refresh:
      i didn't get working the component to handle the WM_DEVICECHANGE message itself and this message didn't also seem
      to appear in the application's message queue, it seems only be posted to the main window, so i had either to
      subclass the parent forms' wndproc (what i didn't like to do) or do it like this:

      to make the autorefresh feature work, you have to declare a message handler in the main form's type declaration:

      type TForm1 = class(TForm);
      ...
        procedure WMDeviceChange(Var Msg:TMessage); message WM_DEVICECHANGE;
      ...
      procedure TForm1.WMDeviceChange(Var Msg:TMessage);
      begin
        if MPDrivePanel1.HandleDeviceChange(Msg) // result is true if drive(s) removed or added
        then
          ... //do something (or not)
        inherited;
      end;}

  published
    { Published declarations }
    property Flat : Boolean read FFlat write SetFlat default True;
    property ButtonWidth : Integer read FButtonWidth write SetButtonWidth nodefault;
    property ButtonHeight : Integer read FButtonHeight write SetButtonHeight nodefault;
    property ButtonSpacing : Integer read FButtonSpacing write SetButtonSpacing default 2;
//    property LargeIcons : Boolean read FLargeIcons write SetLargeIcons default False;
    property ButtonHintTemplate : string read FButtonHintTemplate write SetButtonHintTemplate;
    property ButtonCaptionTemplate : string read FButtonCaptionTemplate write SetButtonCaptionTemplate;
    property ShowCaptions:Boolean read FShowCaptions write SetShowCaptions default True;
    property ShowGlyphs:Boolean read FShowGlyphs write SetShowGlyphs default True;
    property Vertical : Boolean read FVertical write SetVertical default False;
    property Align: TAlign read FAlign write SetAlign;
    property AutoSize: Boolean read FAutoSize write SetAutoSize default True;
    property AllowAllUp : Boolean read FAllowAllUp write SetAllowAllUp default False;
    property GroupIndex : Integer read FGroupIndex write SetGroupIndex default 0;
    property Drive : Char read GetDrive write SetDrive default 'C';

    property OnClickButton : TonClickButton read FOnClickButton write FOnClickButton;
    property Alignment;
    property Anchors;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BiDiMode;
    property BorderWidth;
    property BorderStyle;
    property Color;
    property Constraints;
    property Ctl3D;
    property UseDockManager default True;
    property DockSite;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property FullRepaint;
    property Font;
    property Locked;
    property ParentBiDiMode;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

procedure Register;

function DriveExists ( const ADrive : Char ) : Boolean; // true, if drive ADrive does exist

implementation

uses
  ShellAPI,commctrl;

var
  sfi : TSHFileInfo; // temporary _SHFILEINFOA for shell imagelist's drawing and getting the drive's description
  FBits : set of 0..25; // contains a bitmask of all existing drives

procedure Register;
begin
  RegisterComponents('Merkes'' Pages', [TMPDrivePanel]);
end;

function DriveExists ( const ADrive : Char ) : Boolean;
begin
  Result := (Ord(UpCase(ADrive))-Ord('A')) in FBits;
end;

{ TMPDrivePanel }

constructor TMPDrivePanel.Create(AOwner: TComponent);
begin
  inherited Create (AOwner);
  ControlStyle := ControlStyle-[csSetCaption]; // no caption on the panel
  Integer(FBits) := GetLogicalDrives; // initialize the drive bit mask
  FFlat := True;
  FAutoSize := True;
  FGroupIndex := 0;
  FAllowAllUp := False;
  FShowGlyphs := True;
  FShowCaptions := True;
  inherited AutoSize := False; // we do our own autosizing
  FButtonWidth := GetSystemMetrics(SM_CXSMICON)+26; // width of a small icon + 26 (for the caption)
  FButtonHeight := GetSystemMetrics(SM_CYSMICON)+4;  // height of a small icon +4
  FButtonHintTemplate := ('Click the Button to access Drive "[DD]".');
  FButtonCaptionTemplate := ('[DR]:');
  fVertical := False;
  fButtonSpacing := 2; // space between buttons/buttons and panel border/buttons
//  FLargeIcons := False;
  Caption := '';
  FDrive := 'C';
  inherited Align := alTop;
  FAlign := alTop;
  Height := FButtonHeight+(2*fButtonSpacing);
  ZeroMemory(@sfi, SizeOf(sfi)); // initialize the temporary _SHFILEINFOA
  // get the shell's image list's handle (for small icons)
  FSHIl := SHGetFileInfo(PCHar(ParamStr(0)),0,sfi,SizeOf(sfi),SHGFI_SYSICONINDEX or SHGFI_SMALLICON or SHGFI_ICON);
end;

procedure TMPDrivePanel.CreateWnd;
begin
  inherited CreateWnd ;
  if csDesigning in ComponentState
  then
    DoRefresh;
end;

destructor TMPDrivePanel.Destroy;
var
  i : Integer;
begin
  if Bool(ComponentCount)
  then
    for i := Pred(ComponentCount) downto 0
    do
      if Components[i] is TSpeedButton
      then
        TSpeedButton(Components[i]).Free;

  inherited Destroy;
end;

procedure TMPDrivePanel.DoBtnClick(Sender: TObject);
begin
  FDrive := Char(TSpeedButton(Sender).Tag);
  if Assigned(FOnClickButton)
  then
    FOnClickButton(Self, Sender, Char(TSpeedButton(Sender).Tag));
end;

function TMPDrivePanel.GetDrive: Char;
begin
  Result := #0;
  // can only be used if the buttons are grouped (groupindex <> 0)
  if FGroupIndex = 0
  then
    Exit;
  if ComponentCount = 0
  then
    Exit;
  Result := FDrive;
end;

function TMPDrivePanel.HandleDeviceChange(var Msg: TMessage): Boolean;
var
  i : Integer;
begin
  i := ComponentCount;
  with Msg do
  begin
    if (Msg=WM_DEVICECHANGE) and ((wparam = DBT_DEVICEARRIVAL) or (wparam = DBT_DEVICEREMOVECOMPLETE)) then
    begin
      Integer(FBits) := GetLogicalDrives;
      DoRefresh;
    end;

  end;
  //check if the number of buttons has changed (if so, drives must have been added or removed)
  Result := ComponentCount <> i;

end;

procedure TMPDrivePanel.Loaded;
begin
  inherited Loaded;
  HandleNeeded;
  DoRefresh;
end;

// replace the template marks [DR] and [DD] with the drive char / description
function TMPDrivePanel.ParseDS(const aDrive: Char;
  const Templ: string): string;
var
  i : Integer;
begin
  Result := Templ;
  i := Pos('[DR]',Result);
  while i > 0 do
  begin
    Delete(Result,i,4);
    Insert(UpCase(ADrive),Result,i);
    i := Pos('[DR]',Result);
  end;
  i := Pos('[DD]',Result);
  while i > 0 do
  begin
    SHGetFileInfo(PChar(ADrive+':\'),0,sfi,SizeOf(sfi),SHGFI_DISPLAYNAME);
    Delete(Result,i,4);
    Insert(sfi.szDisplayName,Result,i);
    i := Pos('[DD]',Result);
  end;
end;

procedure TMPDrivePanel.DoRefresh;
var
  i : Integer;
  c : Char;
  l : Integer;
  sfi : TSHFileInfo;
  HasDown : Boolean;
begin
  HasDown := False;
  // remove all buttons
  if Bool(ComponentCount)
  then
    for i := Pred(ComponentCount) downto 0
    do
      if Components[i] is TSpeedButton
      then
        TSpeedButton(Components[i]).Free;

  // check if resizing is needed
  if FAutoSize
  then
    case FAlign of
      alTop,alBottom:
      begin
        Height := (2*fButtonSpacing)+fButtonHeight;
      end;
      alLeft,alRight:
      begin
        Width := (2*fButtonSpacing)+fButtonWidth;
      end;
    end;

  l := fButtonSpacing;
  // create buttons for all existing drives
  for c := 'A' to 'Z'
  do
    if DriveExists(c)
    then
      with TSpeedButton.Create(self) do
      begin
        Parent := self;
        if FShowCaptions
        then
          Caption := ParseDS(c,FButtonCaptionTemplate)
        else
          Caption :='';
        Hint := ParseDS(c,FButtonHintTemplate);
        Flat := FFlat;
        Width := fButtonWidth;
        Height := fButtonHeight;
        Tag := Ord(c); // store the represented drive in the tag property of each button
        AllowAllUp := FAllowAllUp;
        GroupIndex := FGroupIndex;
        if (GroupIndex <> 0) and (FDrive = c) then
        begin
          Down := True;
          HasDown := True; // ok, one button is down (to set the drive property)
        end;
        OnClick := DoBtnClick;
        if not fVertical then
        begin
          Left := l;
          Top := fButtonSpacing;
          Inc(l, fButtonWidth+fButtonSpacing);
        end
        else
        begin
          Top := l;
          Left := fButtonSpacing;
          Inc(l, fButtonHeight+fButtonSpacing);
        end;
        if FShowGlyphs
        then
          with Glyph do
          begin
            Width := GetSystemMetrics(SM_CXSMICON); // width of a small icon
            Height := GetSystemMetrics(SM_CYSMICON); // height of a small icon
            // get the system's icon index
            SHGetFileInfo(PChar(c+':\'),0,sfi,SizeOf(sfi),SHGFI_SYSICONINDEX or SHGFI_SMALLICON or SHGFI_ICON);
            // draw the system icon
            ImageList_Draw(FSHIl,sfi.iIcon,Canvas.Handle,0,0,ILD_NORMAL);
          end;
      end;

  if (not HasDown) and (ComponentCount >0) and (FGroupIndex <>0) and (not FAllowAllUp)
  then
    // before the last refresh a drive was set that isn't existing anymore, so set the current drive
    SetDrive((GetCurrentDir+'C')[1]);
end;

procedure TMPDrivePanel.SetAlign(const Value: TAlign);
var
  ToCH : Boolean;
begin
  ToCh := False;
  if ((FAlign in [alTop,alBottom]) and (Value in [alLeft,alRight])) or
     ((Value in [alTop,alBottom]) and (FAlign in [alLeft,alRight]))
  then
    ToCh := True;

  FAlign := Value;
  Inherited Align := Value;
  if ToCH then
  begin
    FVertical := (FAlign in [alLeft,alRight]);
    DoRefresh;
  end;
end;

procedure TMPDrivePanel.SetAllowAllUp(const Value: Boolean);
begin
  FAllowAllUp := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetAutoSize(const Value: Boolean);
begin
  FAutoSize := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetButtonCaptionTemplate(const Value: string);
begin
  FButtonCaptionTemplate := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetButtonHeight(const Value: Integer);
begin
  FButtonHeight := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetButtonHintTemplate(const Value: string);
begin
  FButtonHintTemplate := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetButtonSpacing(const Value: Integer);
begin
  FButtonSpacing := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetButtonWidth(const Value: Integer);
begin
  FButtonWidth := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetDrive(const Value: Char);
var
  i : Integer;
  r : Char;
begin
  r := #0;
  if FGroupIndex = 0
  then
    Exit;
  if ComponentCount = 0
  then
    Raise Exception.Create('No drive button exists');

  for i:= 0 to Pred(ComponentCount)
  do
    with TSpeedButton(Components[i])
    do
      // push down the corresponding button
      Down := Tag = Ord(UpCase(Value));

  for i:= 0 to Pred(ComponentCount)
  do
    with TSpeedButton(Components[i])
    do
      if Down
      then
      begin
        r := UpCase(Char(Tag));
        if Assigned(FOnClickButton)
        then
          // also fire an onclickbutton event on setting the drive
          FOnClickButton(Self, TSpeedButton(self.Components[i]), Char(Tag));
      end;

  if r = #0
  then
    Raise Exception.Create('Invalid drive');

  FDrive := r;
end;

procedure TMPDrivePanel.SetFlat(const Value: Boolean);
begin
  FFlat := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetGroupIndex(const Value: Integer);
begin
  FGroupIndex := Value;
  DoRefresh;
end;

(*procedure TMPDrivePanel.SetLargeIcons(const Value: Boolean);
begin
  FLargeIcons := Value;
  DoRefresh;
end;*)


procedure TMPDrivePanel.SetShowCaptions(const Value: Boolean);
begin
  FShowCaptions := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetShowGlyphs(const Value: Boolean);
begin
  FShowGlyphs := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.SetVertical(const Value: Boolean);
begin
  FVertical := Value;
  DoRefresh;
end;

procedure TMPDrivePanel.RefreshButtons;
begin
  // rebuild the bit mask of existing drives
  Integer(FBits) := GetLogicalDrives;
  DoRefresh;
end;

end.
